/*
   +-------------------------------------+
   | File: gldraw.c                      |
   |                                     |
   | Copyright 1999 by Michael B. Martin |
   +-------------------------------------+
   | Cubist: Rubik's Cube (tm) simulator |
   | version 1.0                         |
   +-------------------------------------+
*/

#include <GL/gl.h>
#include <GL/glu.h>
#include "cube.h"

// private function prototypes
void thread_cleanup(void);
status_t ThreadEntry(void *p);
void new_thread(void (*func)(), char *str);
void draw_str(char *str, int x, int y);
void enable_lighting(bool enable);
void GL_init(void);
void setColor(int color);
void draw_one_sided_cube(int color1);
void draw_two_sided_cube(int color1, int color2);
void draw_three_sided_cube(int color1, int color2, int color3);
void draw_front_cube_section(void);
void draw_middle_cube_section(void);
void draw_back_cube_section(void);
void draw_cube(void);
void draw_HUD(void);
void display(void);
void reshape(int w, int h);
void report(int value);
void idle(void);
void visible(int vis);
void key(unsigned char key, int x, int y);
void skey(int key, int x, int y);

// material values for the different faces
const static GLfloat cube_mat_red[4] = {0.9, 0.0, 0.0, 1.0};
const static GLfloat cube_mat_yellow[] = {0.9, 0.9, 0.0, 1.0};
const static GLfloat cube_mat_orange[] = {0.9, 0.4, 0.0, 1.0};
const static GLfloat cube_mat_white[] = {1.0, 1.0, 1.0, 1.0};
const static GLfloat cube_mat_blue[] = {0.0, 0.0, 1.0, 1.0};
const static GLfloat cube_mat_green[] = {0.0, 0.6, 0.0, 1.0};
const static GLfloat cube_mat_black[] = {0.0, 0.0, 0.0, 1.0};

// lighting properties
const static GLfloat light0_position[] = {1.0, 1.0, 2.0, 0.0};
const static GLfloat light0_diffuse[] = {0.5, 1.0, 1.0, 1.0};

#define HALF_CUBE 0.5

#define EPS 0.00001

// cube vertices
const static GLfloat cube_vertices[] = 
{
	-HALF_CUBE, -HALF_CUBE, -HALF_CUBE,
	 HALF_CUBE, -HALF_CUBE, -HALF_CUBE,
	 HALF_CUBE,  HALF_CUBE, -HALF_CUBE,
	-HALF_CUBE,  HALF_CUBE, -HALF_CUBE,
	-HALF_CUBE, -HALF_CUBE,  HALF_CUBE,
	 HALF_CUBE, -HALF_CUBE,  HALF_CUBE,
	 HALF_CUBE,  HALF_CUBE,  HALF_CUBE,
	-HALF_CUBE,  HALF_CUBE,  HALF_CUBE,
};

// cube face vertex index arrays
const static GLubyte cubeFrontIndices[4] = {4, 5, 6, 7};
const static GLubyte cubeTopIndices[4] = {2, 3, 7, 6};
const static GLubyte cubeLeftIndices[4] = {0, 4, 7, 3};
const static GLubyte cubeBackIndices[4] = {0, 3, 2, 1};
const static GLubyte cubeBottomIndices[4] = {0, 1, 5, 4};
const static GLubyte cubeRightIndices[4] = {1, 2, 6, 5};

// various flags
bool lighting;		// use OpenGL lighting?
bool rotating;		// apply free rotation?
bool dirty;			// do we need to redraw?
bool animating;		// are we animating a move now?
bool animation_done;// did an animation just finish?
bool moving;		// is a movement thread running?
bool showFPS;		// show frame rate counter in "HUD"?
bool showMessages;	// show "HUD" messages?
bool have_sem;		// does drawing thread have the semaphore?

int frames;				// variables for calculating frame rates
float timeStart, timeEnd, fps;	//    "
sem_id draw_sem;		// semaphore for animating processes
thread_id anim_thread;	// thread for animating processes
GLfloat Xrot, Yrot;		// cube rotation angles
char *message;			// string for informational messages
GLfloat scale_factor;	// scaling factor to adjust cube size

// rotation angles for individual cube sections (for animations)
GLfloat left_right_rot[3];
GLfloat up_down_rot[3];
GLfloat cw_ccw_rot[3];
GLfloat rot_inc;

GLfloat rot_angle[10] = {
	2,3,5,6,9,10,15,18,30,45};	// nontrivial divisors of 90
GLfloat sizes[10] = {0.2,0.5,1.0,1.5,2.0};		// scale factor choices
int rot_index = 5;
int size_index = 2;

// Initialize non-OpenGL stuff
void init(void)
{
	Reset();		// initialize the cube itself

	lighting = false;
	rotating = false;
//	rotating = true;
	dirty = false;
	animating = false;
	animation_done = false;
	moving = false;
	have_sem = false;
	Xrot = Yrot = 0.0;
	left_right_rot[0] = 0.0;
	left_right_rot[1] = 0.0;
	left_right_rot[2] = 0.0;
	up_down_rot[0] = 0.0;
	up_down_rot[1] = 0.0;
	up_down_rot[2] = 0.0;
	cw_ccw_rot[0] = 0.0;
	cw_ccw_rot[1] = 0.0;
	cw_ccw_rot[2] = 0.0;
	rot_inc = rot_angle[rot_index];
	scale_factor = sizes[size_index];
	message = "";
	draw_sem = create_sem(1,"draw_sem");
	if (draw_sem < B_OK)
		printf("Uh oh, couldn't allocate the semaphore!\n");
}

void reinit(void)
{
	rot_inc = rot_angle[rot_index];
	scale_factor = sizes[size_index];
}

void cleanup(void)
{
	if (moving)
		thread_cleanup();
	delete_sem(draw_sem);
}

void showInfo(void)
{
	const GLubyte *str;

	printf("OpenGL driver info:\n");
	str = glGetString(GL_VENDOR);
	printf("  Vendor: \"%s\"\n",str);
	str = glGetString(GL_RENDERER);
	printf("  Renderer: \"%s\"\n",str);
	str = glGetString(GL_VERSION);
	printf("  Version: \"%s\"\n",str);
	str = glGetString(GL_EXTENSIONS);
	printf("  Extensions: \"%s\"\n",str);
}

// Initialize OpenGL state.
void GL_init(void)
{
//	showInfo();
	
	// draw background in dark blue
	glClearColor(0.0, 0.2, 0.4, 0.0);
	// draw background in black
	glClearColor(0.0, 0.0, 0.0, 0.0);

	// add some lighting
	enable_lighting(true);
//	enable_lighting(false);

	// use depth buffering for hidden surface elimination
	glEnable(GL_DEPTH_TEST);

	// use back-face culling to remove some unseen polys
	glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);

//	glEnable(GL_NORMALIZE);

	glShadeModel(GL_SMOOTH);
//	glShadeModel(GL_FLAT);
}

// Tidy up after the thread finishes.
void thread_cleanup(void)
{
	status_t ret;

	message = "";
	moving = false;
//	printf("Cleaning up thread\n");
	kill_thread(anim_thread);
	wait_for_thread(anim_thread,&ret);
}

// Entry point for the new thread.
status_t ThreadEntry(void *p)
{
	void (*func)(void);

	func = p;
//	printf("New thread starting...\n");
	(*func)();
//	printf("Thread done!\n");
	thread_cleanup();
	return(B_OK);		// never returns?
}

// Start a new thread to run the given function.
void new_thread(void (*func)(void), char *str)
{

	if (!moving)
	{
		moving = true;
		anim_thread = spawn_thread(ThreadEntry,"tick_thread",
			B_NORMAL_PRIORITY,(void *)func);
		resume_thread(anim_thread);
//		printf("New thread: %s\n",str);
//		printf("Thread has id %ld, semaphore has id %ld\n",anim_thread,draw_sem);
		if (*str)
			message = str;
	}
	else
	{
		printf("One thread already running, new thread request ignored!\n");
	}
}

// Toggle use of OpenGL lighting.
void enable_lighting(bool enable)
{
	if (enable == true)
	{
		glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
//		glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
		glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0);
		lighting = true;
	}
	else
	{
		glDisable(GL_LIGHTING);
		lighting = false;
	}
}

// Make given color the active color; uses material properties when lighting is
// turned on, plain colors otherwise.
void setColor(int color)
{
	if (lighting)
	{
		switch(color)
		{
			case RED:
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, cube_mat_red);
				break;
			case YELLOW:
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, cube_mat_yellow);
				break;
			case ORANGE:
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, cube_mat_orange);
				break;
			case WHITE:
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, cube_mat_white);
				break;
			case BLUE:
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, cube_mat_blue);
				break;
			case GREEN:
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, cube_mat_green);
				break;
			case BLACK:
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, cube_mat_black);
				break;
			default:
				printf("Unknown color specified!\n");
		}
	}
	else	// no lighting
	{
		switch(color)
		{
			case RED:
				glColor3fv(cube_mat_red);
				break;
			case YELLOW:
				glColor3fv(cube_mat_yellow);
				break;
			case ORANGE:
				glColor3fv(cube_mat_orange);
				break;
			case WHITE:
				glColor3fv(cube_mat_white);
				break;
			case BLUE:
				glColor3fv(cube_mat_blue);
				break;
			case GREEN:
				glColor3fv(cube_mat_green);
				break;
			case BLACK:
				glColor3fv(cube_mat_black);
				break;
			default:
				printf("Unknown color specified!\n");
		}
	}
}

// Draws a cube with five black sides and one colored side; the colored
// side is the front side.
void draw_one_sided_cube(int color1)
{
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, cube_vertices);
	// draw front face first
	glNormal3f(0.0, 0.0, 1.0);
	setColor(color1);
//	glPushMatrix();
//	glScalef(0.9,0.9,1.005);
//	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeFrontIndices);
//	glPopMatrix();
//	setColor(BLACK);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeFrontIndices);
	// now draw top face
	setColor(BLACK);
	glNormal3f(0.0, 1.0, 0.0);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeTopIndices);
	// now draw left face
	glNormal3f(-1.0, 0.0, 0.0);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeLeftIndices);
	// now draw back face
	glNormal3f(0.0, 0.0, -1.0);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeBackIndices);
	// now draw bottom face
	glNormal3f(0.0, -1.0, 0.0);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeBottomIndices);
	// now draw right face
	glNormal3f(1.0, 0.0, 0.0);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeRightIndices);
	glDisableClientState(GL_VERTEX_ARRAY);
}

// Draws a cube with four black sides and two colored sides; the colored
// sides are front and top (in that order!).
void draw_two_sided_cube(int color1, int color2)
{
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, cube_vertices);
	// draw front face first
	setColor(color1);
	glNormal3f(0.0, 0.0, 1.0);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeFrontIndices);
	// now draw top face
	setColor(color2);
	glNormal3f(0.0, 1.0, 0.0);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeTopIndices);
	// now draw left face
	setColor(BLACK);
	glNormal3f(-1.0, 0.0, 0.0);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeLeftIndices);
	// now draw back face
	glNormal3f(0.0, 0.0, -1.0);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeBackIndices);
	// now draw bottom face
	glNormal3f(0.0, -1.0, 0.0);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeBottomIndices);
	// now draw right face
	glNormal3f(1.0, 0.0, 0.0);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeRightIndices);
	glDisableClientState(GL_VERTEX_ARRAY);
}

// Draws a cube with three black sides and three colored sides; the colored
// sides are front, top, and left side (in that order!).
void draw_three_sided_cube(int color1, int color2, int color3)
{
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, cube_vertices);
	// draw front face first
	setColor(color1);
	glNormal3f(0.0, 0.0, 1.0);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeFrontIndices);
	// now draw top face
	setColor(color2);
	glNormal3f(0.0, 1.0, 0.0);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeTopIndices);
	// now draw left face
	setColor(color3);
	glNormal3f(-1.0, 0.0, 0.0);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeLeftIndices);
	// now draw back face
	setColor(BLACK);
	glNormal3f(0.0, 0.0, -1.0);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeBackIndices);
	// now draw bottom face
	glNormal3f(0.0, -1.0, 0.0);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeBottomIndices);
	// now draw right face
	glNormal3f(1.0, 0.0, 0.0);
	glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, cubeRightIndices);
	glDisableClientState(GL_VERTEX_ARRAY);
}

// Draws the front nine sub-cubes.
void draw_front_cube_section(void)
{
	int x, y;
	GLfloat cx, cy;			// x,y coords of center of a sub-cube

	for (x=0; x<3; x++)
	{
		for (y=0; y<3; y++)
		{
			cy = (GLfloat) 1-x;
			cx = (GLfloat) y-1;
			glPushMatrix();
			switch(x*3+y)
			{
				case 0:
					if (animating)
					{
						glRotatef(cw_ccw_rot[0], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[0], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[0], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, 1.0);
					draw_three_sided_cube(side1[0],side5[6],side4[2]);
					break;
				case 1:
					if (animating)
					{
						glRotatef(cw_ccw_rot[0], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[0], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[1], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, 1.0);
					draw_two_sided_cube(side1[1],side5[7]);
					break;
				case 2:
					if (animating)
					{
						glRotatef(cw_ccw_rot[0], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[0], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[2], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, 1.0);
					glRotatef(90.0, 0.0, 1.0, 0.0);
					draw_three_sided_cube(side2[0],side5[8],side1[2]);
					break;
				case 3:
					if (animating)
					{
						glRotatef(cw_ccw_rot[0], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[1], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[0], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, 1.0);
					glRotatef(90.0, 0.0, 0.0, 1.0);
					draw_two_sided_cube(side1[3],side4[5]);
					break;
				case 4:
					if (animating)
					{
						glRotatef(cw_ccw_rot[0], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[1], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[1], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, 1.0);
					draw_one_sided_cube(side1[4]);
					break;
				case 5:
					if (animating)
					{
						glRotatef(cw_ccw_rot[0], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[1], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[2], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, 1.0);
					glRotatef(-90.0, 0.0, 0.0, 1.0);
					draw_two_sided_cube(side1[5],side2[3]);
					break;
				case 6:
					if (animating)
					{
						glRotatef(cw_ccw_rot[0], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[2], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[0], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, 1.0);
					glRotatef(90.0, 0.0, 0.0, 1.0);
					draw_three_sided_cube(side1[6],side4[8],side6[0]);
					break;
				case 7:
					if (animating)
					{
						glRotatef(cw_ccw_rot[0], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[2], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[1], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, 1.0);
					glRotatef(180.0, 0.0, 0.0, 1.0);
					draw_two_sided_cube(side1[7],side6[1]);
					break;
				case 8:
					if (animating)
					{
						glRotatef(cw_ccw_rot[0], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[2], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[2], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, 1.0);
					glRotatef(90.0, 0.0, 1.0, 0.0);
					glRotatef(90.0, 0.0, 0.0, 1.0);
					draw_three_sided_cube(side2[6],side1[8],side6[2]);
					break;
			}
			glPopMatrix();
		}
	}
}

// Draws the middle eight sub-cubes.
void draw_middle_cube_section(void)
{
	int x, y;
	GLfloat cx, cy;			// x,y coords of center of a sub-cube

	for (x=0; x<3; x++)
	{
		for (y=0; y<3; y++)
		{
			if (x*3+y == 4)
				continue;
			cy = (GLfloat) 1-x;
			cx = (GLfloat) y-1;
			glPushMatrix();
			switch(x*3+y)
			{
				case 0:
					if (animating)
					{
						glRotatef(cw_ccw_rot[1], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[0], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[0], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, 0.0);
					glRotatef(-90.0, 0.0, 1.0, 0.0);
					draw_two_sided_cube(side4[1],side5[3]);
					break;
				case 1:
					if (animating)
					{
						glRotatef(cw_ccw_rot[1], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[0], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[1], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, 0.0);
					glRotatef(-90.0, 1.0, 0.0, 0.0);
					draw_one_sided_cube(side5[4]);
					break;
				case 2:
					if (animating)
					{
						glRotatef(cw_ccw_rot[1], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[0], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[2], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, 0.0);
					glRotatef(90.0, 0.0, 1.0, 0.0);
					draw_two_sided_cube(side2[1],side5[5]);
					break;
				case 3:
					if (animating)
					{
						glRotatef(cw_ccw_rot[1], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[1], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[0], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, 0.0);
					glRotatef(-90.0, 0.0, 1.0, 0.0);
					draw_one_sided_cube(side4[4]);
					break;
				case 5:
					if (animating)
					{
						glRotatef(cw_ccw_rot[1], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[1], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[2], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, 0.0);
					glRotatef(90.0, 0.0, 1.0, 0.0);
					draw_one_sided_cube(side2[4]);
					break;
				case 6:
					if (animating)
					{
						glRotatef(cw_ccw_rot[1], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[2], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[0], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, 0.0);
					glRotatef(180.0, 0.0, 0.0, 1.0);
					glRotatef(90.0, 0.0, 1.0, 0.0);
					draw_two_sided_cube(side4[7],side6[3]);
					break;
				case 7:
					if (animating)
					{
						glRotatef(cw_ccw_rot[1], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[2], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[1], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, 0.0);
					glRotatef(90.0, 1.0, 0.0, 0.0);
					draw_one_sided_cube(side6[4]);
					break;
				case 8:
					if (animating)
					{
						glRotatef(cw_ccw_rot[1], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[2], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[2], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, 0.0);
					glRotatef(180.0, 0.0, 0.0, 1.0);
					glRotatef(-90.0, 0.0, 1.0, 0.0);
					draw_two_sided_cube(side2[7],side6[5]);
					break;
			}
			glPopMatrix();
		}
	}
}

// Draws the front nine sub-cubes.
void draw_back_cube_section(void)
{
	int x, y;
	GLfloat cx, cy;			// x,y coords of center of a sub-cube

	for (x=0; x<3; x++)
	{
		for (y=0; y<3; y++)
		{
			cy = (GLfloat) 1-x;
			cx = (GLfloat) y-1;
			glPushMatrix();
			switch(x*3+y)
			{
				case 0:
					if (animating)
					{
						glRotatef(cw_ccw_rot[2], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[0], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[0], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, -1.0);
					glRotatef(-90.0, 0.0, 1.0, 0.0);
					draw_three_sided_cube(side4[0],side5[0],side3[2]);
					break;
				case 1:
					if (animating)
					{
						glRotatef(cw_ccw_rot[2], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[0], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[1], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, -1.0);
					glRotatef(180.0, 0.0, 1.0, 0.0);
					draw_two_sided_cube(side3[1],side5[1]);
					break;
				case 2:
					if (animating)
					{
						glRotatef(cw_ccw_rot[2], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[0], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[2], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, -1.0);
					glRotatef(180.0, 0.0, 1.0, 0.0);
					draw_three_sided_cube(side3[0],side5[2],side2[2]);
					break;
				case 3:
					if (animating)
					{
						glRotatef(cw_ccw_rot[2], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[1], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[0], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, -1.0);
					glRotatef(180.0, 0.0, 1.0, 0.0);
					glRotatef(-90.0, 0.0, 0.0, 1.0);
					draw_two_sided_cube(side3[5],side4[3]);
					break;
				case 4:
					if (animating)
					{
						glRotatef(cw_ccw_rot[2], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[1], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[1], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, -1.0);
					glRotatef(180.0, 0.0, 1.0, 0.0);
					draw_one_sided_cube(side3[4]);
					break;
				case 5:
					if (animating)
					{
						glRotatef(cw_ccw_rot[2], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[1], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[2], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, -1.0);
					glRotatef(180.0, 0.0, 1.0, 0.0);
					glRotatef(90.0, 0.0, 0.0, 1.0);
					draw_two_sided_cube(side3[3],side2[5]);
					break;
				case 6:
					if (animating)
					{
						glRotatef(cw_ccw_rot[2], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[2], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[0], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, -1.0);
					glRotatef(180.0, 0.0, 1.0, 0.0);
					glRotatef(90.0, 0.0, 1.0, 0.0);
					glRotatef(90.0, 0.0, 0.0, 1.0);
					draw_three_sided_cube(side4[6],side3[8],side6[6]);
					break;
				case 7:
					if (animating)
					{
						glRotatef(cw_ccw_rot[2], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[2], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[1], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, -1.0);
					glRotatef(180.0, 0.0, 1.0, 0.0);
					glRotatef(180.0, 0.0, 0.0, 1.0);
					draw_two_sided_cube(side3[7],side6[7]);
					break;
				case 8:
					if (animating)
					{
						glRotatef(cw_ccw_rot[2], 0.0, 0.0, 1.0);
						glRotatef(left_right_rot[2], 0.0, 1.0, 0.0);
						glRotatef(up_down_rot[2], 1.0, 0.0, 0.0);
					}
					glTranslatef(cx, cy, -1.0);
					glRotatef(180.0, 0.0, 1.0, 0.0);
					glRotatef(90.0, 0.0, 0.0, 1.0);
					draw_three_sided_cube(side3[6],side2[8],side6[8]);
					break;
			}
			glPopMatrix();
		}
	}
}

// Renders the Rubik's Cube in the OpenGL framebuffer.
void draw_cube(void)
{
/*
	int i, j;

	for(i=0; i<3; i++)
	{
		for(j=0; j<3; j++)
			switch(side1[3*i+j])
			{
				case RED:
					printf("R ");
					break;
				case YELLOW:
					printf("Y ");
					break;
				case ORANGE:
					printf("O ");
					break;
				case WHITE:
					printf("W ");
					break;
				case BLUE:
					printf("B ");
					break;
				case GREEN:
					printf("G ");
					break;
				default:
					break;
			}
		printf("\n");
	}
	printf("\n");
*/

//	glEnable(GL_NORMALIZE);	// fix normals for lighting when rescaling
	glScalef(scale_factor,scale_factor,scale_factor);
	draw_front_cube_section();
	draw_middle_cube_section();
	draw_back_cube_section();
//	glDisable(GL_NORMALIZE);
}

// Display callback that draws each frame.
void drawScene(void)
{
	int i;

//	printf("drawScene() called (have_sem = %d)\n",have_sem);

	if (!have_sem)
	{
//		printf("Trying to acquire semaphore in drawScene()...\n");
		if (acquire_sem(draw_sem) != B_NO_ERROR) {
			printf("Error acquiring semaphore (%ld)\n",draw_sem);
			return;		// give up
		}
		else
		{
			have_sem = true;
//			printf("acquired semaphore in drawScene()\n");
		}
	}

	// reset buffers
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// set camera view
	glLoadIdentity();
	glTranslatef(0.0, 0.0, -15.0);
	glRotatef(15.0, 1.0, 0.0, 0.0);
	glRotatef(-15.0, 0.0, 1.0, 0.0);

	glRotatef(Xrot, 1.0, 0.0, 0.0);
	glRotatef(Yrot, 0.0, 1.0, 0.0);

//	glEnable(GL_LINE_SMOOTH);

	if (1)
		draw_cube();
	else
	{
		// use polygon offset; will be needed later
		draw_cube();
		glEnable(GL_POLYGON_OFFSET_LINE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glPolygonOffset(-1.0, -1.0);
		glLineWidth(2.0);
		draw_cube();
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glDisable(GL_POLYGON_OFFSET_LINE);
	}

//	draw_HUD();		// uncomment this for the GLUT version

    if(glGetError()) /* to catch programming errors; should never happen */
       printf("Oops! I screwed up my OpenGL calls somewhere\n");

	// sync and update screen
	glFlush();

	if (dirty)
		dirty = false;

	if (animation_done)		// don't release unless an animation just finished!
	{
		for (i=0; i<3; i++)
		{
			if (ABS(left_right_rot[i]) > 89.0)
			{
				left_right_rot[i] = 0.0;
			}
			if (ABS(up_down_rot[i]) > 89.0)
			{
				up_down_rot[i] = 0.0;
			}
			if (ABS(cw_ccw_rot[i]) > 89.0)
			{
				cw_ccw_rot[i] = 0.0;
			}
		}

		animating = false;
		animation_done = false;
		have_sem = false;
		release_sem(draw_sem);
//		printf("semaphore released in drawScene()\n");
	}
}

// Window reshaping callback.
void reshape(int w, int h)
{
	GLdouble aspect;

	// reset the viewing paramters
	aspect = ((GLdouble) w)/((GLdouble) h);
    glViewport(0, 0, w, h);
    glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(40.0, aspect, 1.0, 40.0);
	glMatrixMode(GL_MODELVIEW);
}

// changes to perform between frames
void update(void)
{
	int i;

	if (rotating)
	{
		Xrot += 0.20*10.0;
		Yrot += 0.10*10.0;
		dirty = true;
	}

	// update animation parameters
	if (animating)
	{
//		printf("animation update:\n");
		for (i=0; i<3; i++)
		{
			if (left_right_rot[i] > 0.0)
				left_right_rot[i] += rot_inc;
			else if (left_right_rot[i] < 0.0)
				left_right_rot[i] -= rot_inc;
			if (ABS(left_right_rot[i]) > 89.0)
			{
				animation_done = true;
			}
		}
		for (i=0; i<3; i++)
		{
//		printf("up_down_rot[0] = %.6f\n",up_down_rot[i]);
			if (up_down_rot[i] > 0.0)
				up_down_rot[i] += rot_inc;
			else if (up_down_rot[i] < 0.0)
				up_down_rot[i] -= rot_inc;
			if (ABS(up_down_rot[i]) > 89.0)
			{
				animation_done = true;
			}
		}
		for (i=0; i<3; i++)
		{
			if (cw_ccw_rot[i] > 0.0)
				cw_ccw_rot[i] += rot_inc;
			else if (cw_ccw_rot[i] < 0.0)
				cw_ccw_rot[i] -= rot_inc;
			if (ABS(cw_ccw_rot[i]) > 89.0)
			{
				animation_done = true;
			}
		}
/*
		if (animation_done)
		{
			printf("animation finished\n");
		}
*/
		dirty = true;
//		printf("Setting \"dirty\" flag in update()\n");

//		printf("up_down_rot[0] = %.2f\n",up_down_rot[0]);
	}
}

// Drawing callback called by functions in cube.c; used to indicate when a
// movement function should be displayed.
void draw(int move)
{
//	dirty = true;
//	printf("Setting \"dirty\" flag in draw()\n");

//	printf("draw() called with move==%d (animating = %d)\n",move,animating);

	switch(move)
	{
		case DONE:
			release_sem(draw_sem);
//			printf("semaphore released in draw() (animating = %d)\n",animating);
			if (0)
			{
				int32 thread_count;
				get_sem_count(draw_sem, &thread_count);
				printf("semaphore count is %ld, with moves = %d\n",thread_count,moves);
			}
			return;
			break;
		case TOP_LEFT:
			left_right_rot[0] = -EPS;
			animating = true;
			break;
		case CENTER_LEFT:
			left_right_rot[1] = -EPS;
			animating = true;
			break;
		case BOTTOM_LEFT:
			left_right_rot[2] = -EPS;
			animating = true;
			break;
		case TOP_RIGHT:
			left_right_rot[0] = EPS;
			animating = true;
			break;
		case CENTER_RIGHT:
			left_right_rot[1] = EPS;
			animating = true;
			break;
		case BOTTOM_RIGHT:
			left_right_rot[2] = EPS;
			animating = true;
			break;
		case LEFT_UP:
			up_down_rot[0] = -EPS;
			animating = true;
			break;
		case CENTER_UP:
			up_down_rot[1] = -EPS;
			animating = true;
			break;
		case RIGHT_UP:
			up_down_rot[2] = -EPS;
			animating = true;
			break;
		case LEFT_DOWN:
			up_down_rot[0] = EPS;
			animating = true;
			break;
		case CENTER_DOWN:
			up_down_rot[1] = EPS;
			animating = true;
			break;
		case RIGHT_DOWN:
			up_down_rot[2] = EPS;
			animating = true;
			break;
		case FRONT_CLOCKWISE:
			cw_ccw_rot[0] = -EPS;
			animating = true;
			break;
		case FRONT_COUNTERCLOCKWISE:
			cw_ccw_rot[0] = EPS;
			animating = true;
			break;
		case UP:
			up_down_rot[0] = -EPS;
			up_down_rot[1] = -EPS;
			up_down_rot[2] = -EPS;
			animating = true;
			break;
		case DOWN:
			up_down_rot[0] = EPS;
			up_down_rot[1] = EPS;
			up_down_rot[2] = EPS;
			animating = true;
			break;
		case LEFT:
			left_right_rot[0] = -EPS;
			left_right_rot[1] = -EPS;
			left_right_rot[2] = -EPS;
			animating = true;
			break;
		case RIGHT:
			left_right_rot[0] = EPS;
			left_right_rot[1] = EPS;
			left_right_rot[2] = EPS;
			animating = true;
			break;
		case CLOCKWISE:
			cw_ccw_rot[0] = -EPS;
			cw_ccw_rot[1] = -EPS;
			cw_ccw_rot[2] = -EPS;
			animating = true;
			break;
		case COUNTERCLOCKWISE:
			cw_ccw_rot[0] = EPS;
			cw_ccw_rot[1] = EPS;
			cw_ccw_rot[2] = EPS;
			animating = true;
			break;
		default:
			printf("Unknown movement type (%d) !\n",move);
			return;
	}

	while (animating)
		snooze(2000);	// yield the CPU while waiting

	if (0)
	{
		int32 thread_count;
		get_sem_count(draw_sem, &thread_count);
		printf("semaphore count is %ld, with moves = %d\n",thread_count,moves);
	}

//	printf("Trying to acquire semaphore in draw()...\n");
	if (acquire_sem(draw_sem) != B_NO_ERROR) {
		printf("Error acquiring semaphore (%ld)\n",draw_sem);
	}
	else
	{
//		printf("acquired semaphore in draw() (move==%d)\n",move);
	}
}
